from . import leaflet  # noqa: F401
try:
    from . import mpl  # noqa: F401
    WITH_CARTOPY = True
except ImportError:  # pragma: no cover
    WITH_CARTOPY = False
from .base import Map  # noqa: F401



import json
import string

import yattag
from clldutils import svg

from .base import Map
import cldfviz


BASE_LAYERS = {
    'OpenStreetMap': (
        'http://{s}.tiles.wmflabs.org/bw-mapnik/{z}/{x}/{y}.png',
        {
            'maxZoom': 18,
            'attribution':
                '&copy; <a href="http://www.openstreetmap.org/copyright">OpenStreetMap</a>'}
    ),
    'OpenTopoMap': (
        'https://{s}.tile.opentopomap.org/{z}/{x}/{y}.png',
        {
            'maxZoom': 17,
            'attribution':
                'Map data: &copy; <a href="https://www.openstreetmap.org/copyright">OpenStreetMap'
                '</a> contributors, <a href="http://viewfinderpanoramas.org">SRTM</a> | '
                'Map style: &copy; <a href="https://opentopomap.org">OpenTopoMap</a> '
                '(<a href="https://creativecommons.org/licenses/by-sa/3.0/">CC-BY-SA</a>)'}
    ),
    'Esri_WorldImagery': (
        'https://server.arcgisonline.com/ArcGIS/rest/services/World_Imagery/MapServer/'
        'tile/{z}/{y}/{x}',
        {
            'attribution':
                'Tiles &copy; Esri &mdash; Source: Esri, i-cubed, USDA, USGS, AEX, GeoEye, '
                'Getmapping, Aerogrid, IGN, IGP, UPR-EGP, and the GIS User Community'}
    ),
    'Esri_WorldPhysical': (
        'https://server.arcgisonline.com/ArcGIS/rest/services/World_Physical_Map/MapServer/'
        'tile/{z}/{y}/{x}',
        {
            'attribution': 'Tiles &copy; Esri &mdash; Source: US National Park Service',
            'maxZoom': 8}
    ),
    'Esri_NatGeoWorldMap': (
        'https://server.arcgisonline.com/ArcGIS/rest/services/NatGeo_World_Map/MapServer/'
        'tile/{z}/{y}/{x}',
        {
            'attribution':
                'Tiles &copy; Esri &mdash; National Geographic, Esri, DeLorme, NAVTEQ, UNEP-WCMC, '
                'USGS, NASA, ESA, METI, NRCAN, GEBCO, NOAA, iPC',
            'maxZoom': 16}
    ),
}


class MapLeaflet(Map):
    __formats__ = ['html']

    def __init__(self, languages, args):
        Map.__init__(self, languages, args)
        self.features = []
        self.legend = ''

    @staticmethod
    def add_options(parser, help_suffix):
        parser.add_argument(
            '--base-layer',
            default='OpenStreetMap',
            help="Tile layer for Leaflet maps. {}".format(help_suffix),
            choices=list(BASE_LAYERS),
        )

    def _lonlat(self, language):
        lon, lat = language.lon, language.lat
        if self.args.pacific_centered and lon <= -26:
            lon += 360  # make the map pacific-centered.
        return [lon, lat]

    def add_language(self, language, values, colormaps):
        self.features.append({
            # A language as GeoJSON point with svg marker icon
            "geometry": {"coordinates": self._lonlat(language), "type": "Point"},
            "id": language.id,
            "properties": {
                "name": language.name,
                "values": ' / '.join(['{}'.format(vals[0].v) for vals in values.values() if vals]),
                "icon": svg.data_url(svg.pie(
                    [1] * len(values),
                    [colormaps[pid](vals[0].v) for pid, vals in values.items()],
                    stroke_circle=True)),
                "markerSize": self.args.markersize,
            },
            "type": "Feature"
        })

    def add_legend(self, parameters, colormaps):
        doc, tag, text = yattag.Doc().tagtext()

        def marker(colors):
            doc.stag('img', src=svg.data_url(svg.pie(
                [1] * len(parameters),
                colors,
                stroke_circle=True,
            )), width="{}".format(min([20, self.args.markersize * 2])))

        with tag('table', klass="legend"):
            for i, (pid, parameter) in enumerate(parameters.items()):
                if i != 0:
                    with tag('tr'):
                        with tag('th', colspan=2):
                            doc.stag('hr')
                with tag('tr'):
                    with tag('th'):
                        marker(['#000000' if j == i else '#ffffff' for j in range(len(parameters))])
                    with tag('th', style="text-align: left;"):
                        text(parameter.name)
                if isinstance(parameter.domain, tuple):
                    min_, max_ = parameter.domain
                    with tag('tr'):
                        with tag('td', colspan=2):
                            with tag('table'):
                                with tag('tr'):
                                    with tag('td'):
                                        text(str(round(min_, 2)))
                                    for _ in range(9):
                                        with tag('td'):
                                            text(' ')
                                    with tag('td', style="text-align: right;"):
                                        text(str(round(max_, 2)))
                                with tag('tr'):
                                    for j in range(11):
                                        with tag(
                                            'td',
                                            style='height: 20px; width: 1em; background-color: '
                                                  '{};'.format(
                                                colormaps[pid](min_ + j * (max_ - min_) / 10))
                                        ):
                                            text(' ')
                else:
                    for v, label in parameter.domain.items():
                        with tag('tr'):
                            with tag('td'):
                                marker([colormaps[pid](v) if j == i else '#ffffff'
                                        for j in range(len(parameters))])
                            with tag('td'):
                                text(str(label))
        self.legend = doc.getvalue()

    def __exit__(self, exc_type, exc_val, exc_tb):
        """write files"""
        html = string.Template(
            cldfviz.PKG_DIR.joinpath('templates', 'map', 'leaflet.html').read_text(encoding='utf8')
        ).substitute(
            title=self.args.title or '',
            legend=self.legend,
            options=json.dumps({'language_labels': self.args.language_labels}),
            tile_url=json.dumps(BASE_LAYERS[self.args.base_layer][0]),
            tile_options=json.dumps(BASE_LAYERS[self.args.base_layer][1]),
            geojson=json.dumps({"features": self.features, "type": "FeatureCollection"}))
        self.args.output.write_text(html, encoding='utf8')



from matplotlib import cm
from matplotlib.colors import Normalize, to_hex
import matplotlib.pyplot as plt
from clldutils.color import qualitative_colors, sequential_colors

from cldfviz.multiparameter import CONTINUOUS, CATEGORICAL

COLORMAPS = {
    CATEGORICAL: ['boynton', 'tol', 'base', 'seq'],
    CONTINUOUS: [cm for cm in plt.colormaps() if not cm.endswith('_r')],
}


class Colormap:

    def __init__(self, domain, name=None, novalue=None):
        self.novalue = novalue
        self._cm = getattr(cm, name or 'yyy', cm.jet)

        if isinstance(domain, tuple):
            # Initialize matplotlib colormap and normalizer:
            norm = Normalize(domain[0], domain[1])
            self.cm = lambda v: to_hex(self._cm(norm(float(v))))
        else:
            if name == 'seq':
                colors = sequential_colors(len(domain))
            else:
                colors = qualitative_colors(len(domain), set=name)
            self.cm = lambda v: dict(zip(domain, colors))[v]

    def scalar_mappable(self):
        return cm.ScalarMappable(norm=None, cmap=self._cm)

    def __call__(self, value):
        if value is None:
            return self.novalue
        return self.cm(value)



import webbrowser

PACIFIC_CENTERED = 154


class Map:
    __formats__ = []

    def __init__(self, languages, args):
        self.languages = languages
        self.args = args

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):  # pragma: no cover
        """write files"""
        raise NotImplementedError()

    @staticmethod
    def add_options(parser, help_suffix):  # pragma: no cover
        pass

    def add_language(self, language, values, colormaps):  # pragma: no cover
        raise NotImplementedError()

    def add_legend(self, parameters, colormaps):  # pragma: no cover
        raise NotImplementedError()

    def open(self):  # pragma: no cover
        webbrowser.open(self.args.output.resolve().as_uri())



"""
Plot values for parameters in a CLDF StructureDataset on a map.

Usage examples:
- plot languages of a dataset:
  cldfbench cldfviz.map PATH/TO/DATASET --language-labels
- plot values of one parameter:
  cldfbench cldfviz.map PATH/TO/DATASET --parameters PID
- plot values of a column in the dataset's LanguageTable:
  cldfbench cldfviz.map PATH/TO/DATASET --language-property Macroarea
"""
import pathlib

from pycldf.cli_util import get_dataset, add_dataset
from clldutils.clilib import PathType
from cldfbench.cli_util import add_catalog_spec

from cldfviz.colormap import Colormap, COLORMAPS
from cldfviz.multiparameter import MultiParameter, CONTINUOUS, CATEGORICAL
from cldfviz.map import Map
from cldfviz.cli_util import add_testable

FORMATS = {}
for cls in Map.__subclasses__():
    for fmt in cls.__formats__:
        FORMATS[fmt] = cls


def join_quoted(items):
    return ', '.join(['"{}"'.format(i) for i in items])


def register(parser):
    add_testable(parser)
    add_dataset(parser)
    add_catalog_spec(parser, 'glottolog')
    parser.add_argument(
        '--parameters',
        default=[],
        type=lambda s: s.split(','),
        help="Comma-separated Parameter IDs, specifying the values to plot on the map. If not "
             "specified, all languages in the dataset will be plotted.",
    )
    parser.add_argument(
        '--colormaps',
        default=[],
        type=lambda s: s.split(','),
        help="Comma-separated names of colormaps to use for the respective parameter. Choose from "
             "{} for categorical and from {} for continuous parameters.".format(
            join_quoted(COLORMAPS[CATEGORICAL]), join_quoted(COLORMAPS[CONTINUOUS])),
    )
    parser.add_argument(
        '--language-property',
        help="Plot values of a language property, i.e. a column in the dataset's LanguageTable.",
        default=None,
    )
    parser.add_argument(
        '--language-property-colormap',
        default='tol',
        help="Name of colormap to use for the language property. See help for --colormaps for "
             "choices",
    )
    parser.add_argument(
        '--output',
        type=PathType(type='file', must_exist=False),
        help="Filesystem path to write the resulting map to. If no suffix is specified, it will "
             "be appended according to FORMAT; if given, suffix must match FORMAT.",
        default=pathlib.Path('map'))
    parser.add_argument(
        '--format',
        default='html',
        metavar='FORMAT',
        choices=list(FORMATS),
    )
    parser.add_argument(
        '--markersize',
        help="Size of map markers in pixels",
        type=int,
        default=10,
    )
    parser.add_argument(
        '--title',
        default=None,
        help="Title for the map plot",
    )
    parser.add_argument(
        '--pacific-centered',
        action='store_true',
        default=False,
        help="Center maps of the whole world at the pacific, thus not cutting large language "
             "families in half."
    )
    parser.add_argument(
        '--language-labels',
        action='store_true',
        default=False,
        help="Display language names on the map",
    )
    for cls in Map.__subclasses__():
        cls.add_options(
            parser, help_suffix='(Only for FORMATs {})'.format(join_quoted(cls.__formats__)))


def run(args):
    ds = get_dataset(args)

    if not args.output.suffix:
        args.output = args.output.parent / "{}.{}".format(args.output.name, args.format)
    else:
        assert args.output.suffix[1:] == args.format

    glottolog = {lg.id: lg for lg in args.glottolog.api.languoids() if lg.latitude is not None} \
        if args.glottolog else {}
    data = MultiParameter(
        ds, args.parameters, glottolog=glottolog, language_property=args.language_property)
    if args.colormaps:
        if args.language_property:
            args.colormaps.append(args.language_property_colormap)
        assert len(args.colormaps) == len(data.parameters)
    else:
        args.colormaps = [None] * len(data.parameters)
    cms = {pid: Colormap(data.parameters[pid].domain, name=cm)
           for pid, cm in zip(data.parameters, args.colormaps)}

    with FORMATS[args.format](data.languages.values(), args) as fig:
        for lang, values in data.iter_languages():
            fig.add_language(lang, values, cms)

        fig.add_legend(data.parameters, cms)

        args.log.info('Output written to: {}'.format(args.output))
        if not args.test:
            fig.open()  # pragma: no cover



import decimal
import functools
import itertools
import collections

import attr

CONTINUOUS = 1
CATEGORICAL = 2


@attr.s
class Language:
    id = attr.ib()
    name = attr.ib()
    lat = attr.ib()
    lon = attr.ib()

    @classmethod
    def from_object(cls, obj, glottolog=None):
        lonlat = obj.lonlat
        glang = (glottolog or {}).get(getattr(obj.cldf, 'glottocode', obj.id))
        if (not lonlat) and glang and glang.latitude is not None:
            lonlat = (glang.longitude, glang.latitude)
        if lonlat:
            return cls(
                id=obj.id,
                name=getattr(obj.cldf, 'name', obj.id),
                lon=float(lonlat[0]),
                lat=float(lonlat[1]))

    @classmethod
    def from_glottolog(cls, lid, glottolog):
        if glottolog:
            glang = glottolog.get(lid)
            if glang and glang.latitude is not None:
                return cls(id=glang.id, name=glang.name, lat=glang.latitude, lon=glang.longitude)


@attr.s
class Parameter:
    id = attr.ib()
    name = attr.ib()
    type = attr.ib(default=CATEGORICAL)
    domain = attr.ib(default=attr.Factory(dict))

    @classmethod
    def from_object(cls, obj):
        return cls(id=obj.id, name=getattr(obj.cldf, 'name', obj.id))


@functools.total_ordering
@attr.s(order=False, eq=False)
class Value:
    v = attr.ib()
    pid = attr.ib()
    lid = attr.ib()
    code = attr.ib()
    float = attr.ib(default=None)

    def __attrs_post_init__(self):
        try:
            self.float = float(self.v)
        except ValueError:
            self.float = None

    def __eq__(self, other):
        return (self.lid, self.pid, self.v) == (other.lid, other.pid, other.v)

    def __lt__(self, other):
        return (self.lid, self.pid, self.v) < (other.lid, other.pid, other.v)

    @classmethod
    def from_row(cls, row, codes):
        return cls(
            v=row.get('codeReference') or row['value'],
            lid=row['languageReference'],
            pid=row['parameterReference'],
            code=codes[row['parameterReference']][row['codeReference']]
            if row['parameterReference'] in codes else None,
        )


class MultiParameter:
    def __init__(self, ds, pids, include_missing=False, glottolog=None, language_property=None):
        #
        # FIXME: If no pids, look for language_col, if none either, make synthetic "has_language"
        # parameter.
        #
        langs = {lg.id: Language.from_object(lg, glottolog=glottolog)
                 for lg in ds.objects('LanguageTable')} if 'LanguageTable' in ds else {}
        params = {p.id: Parameter.from_object(p)
                  for p in ds.objects('ParameterTable')} if 'ParameterTable' in ds else {}
        self.parameters = collections.OrderedDict(
            [(pid, params.get(pid, Parameter(id=pid, name=pid))) for pid in pids])
        if language_property:
            self.parameters[language_property] = Parameter(
                id=language_property, name=language_property)
        elif not pids:
            # No parameters and no language property specified: Just plot language locations.
            self.parameters['language'] = Parameter(id='language', name='language')

        codes = collections.defaultdict(collections.OrderedDict)
        if 'CodeTable' in ds:
            for row in ds.iter_rows('CodeTable', 'id', 'parameterReference', 'name'):
                if row['parameterReference'] in self.parameters:
                    codes[row['parameterReference']][row['id']] = row['name']
        language_rows = []
        if language_property:
            language_rows = list(ds.iter_rows('LanguageTable', 'id', 'name'))
            if not all(isinstance(v[language_property], (int, float, decimal.Decimal))
                       for v in language_rows if v[language_property] is not None):
                codes[language_property] = collections.OrderedDict([
                    (p, p) for p in sorted(
                        set(r[language_property] for r in language_rows
                            if r[language_property] is not None))
                ])
        self.languages = collections.OrderedDict()
        self.values = []
        colmap = ['languageReference', 'parameterReference', 'value']
        if codes:
            colmap.append('codeReference')
        for val in ds.iter_rows('ValueTable', *colmap):
            if (val['value'] is not None) and val['parameterReference'] in self.parameters:
                lang = langs.get(val['languageReference'])
                if not lang:
                    lang = Language.from_glottolog(val['languageReference'], glottolog)
                if lang:
                    self.languages[val['languageReference']] = lang
                    self.values.append(Value.from_row(val, codes))
        if language_property:
            for lang in language_rows:
                if lang[language_property] is not None:
                    self.languages.setdefault(lang['id'], langs[lang['id']])
                    self.values.append(Value(
                        v=lang[language_property],
                        pid=language_property,
                        lid=lang['id'],
                        code=language_property))
        elif not pids:
            # No parameters and no language property specified: Just plot language locations.
            for lang in ds.iter_rows('LanguageTable', 'id', 'name'):
                self.languages.setdefault(lang['id'], langs[lang['id']])
                self.values.append(Value(
                    v='y',
                    pid='language',
                    lid=lang['id'],
                    code='language'))

        for p in self.parameters.values():
            if p.id in codes:
                p.domain = codes[p.id]
            else:
                vals = [v for v in self.values if v.pid == p.id]
                if all(v.float is not None for v in vals) and len(set(v.v for v in vals)) > 8:
                    p.type = CONTINUOUS
                    p.domain = (min(v.float for v in vals), max(v.float for v in vals))
                else:
                    p.domain = collections.OrderedDict([
                        (v, v) for v in sorted(set(vv.v for vv in vals), key=lambda vv: str(vv))])

        self.include_missing = include_missing

    def iter_languages(self):
        for lid, values in itertools.groupby(sorted(self.values), lambda v: v.lid):
            values = {pid: list(vals) for pid, vals in itertools.groupby(values, lambda v: v.pid)}
            values = collections.OrderedDict(
                [(pid, values.get(pid, [])) for pid in self.parameters])
            if self.include_missing or all(bool(v) for v in values.values()):
                yield self.languages[lid], values



"""
Map plotting with matplotlib and cartopy
"""
import textwrap

import cartopy.feature
import cartopy.crs
from matplotlib import pyplot as plt
from matplotlib.patches import Wedge
from matplotlib.legend_handler import HandlerPatch
from PIL import Image

from .base import Map, PACIFIC_CENTERED


class HandleWedge(HandlerPatch):
    def create_artists(
            self,
            legend,
            orig_handle,
            xdescent,
            ydescent,
            width,
            height,
            fontsize,
            trans):
        center = 0.5 * width - 0.5 * xdescent, 0.5 * height - 0.5 * ydescent
        p = Wedge(
            center, height / 1.5, orig_handle.theta1, orig_handle.theta2, width=orig_handle.width)
        self.update_prop(p, orig_handle, legend)
        p.set_transform(trans)
        return [p]


class MapPlot(Map):
    """
    A geographic map, zoomed in to the extent defined by a list of languages.
    """
    __formats__ = ['jpg', 'png', 'pdf']

    def __init__(self, languages, args):
        Map.__init__(self, languages, args)
        lats, lons = [k.lat for k in languages], [k.lon for k in languages]
        self.central_longitude = PACIFIC_CENTERED if args.pacific_centered else 0
        self.extent = [
            round(min(lons) - args.padding_left, 1) + self.central_longitude,
            round(max(lons) + args.padding_right, 1) + self.central_longitude,
            round(min(lats) - args.padding_bottom, 1),
            round(max(lats) + args.padding_top, 1)
        ]
        self.ax = None
        self.scaling_factor = 1
        self.proj = cartopy.crs.PlateCarree(central_longitude=self.central_longitude)

    def __enter__(self):
        plt.clf()
        fig = plt.figure(figsize=(self.args.width, self.args.height), dpi=self.args.dpi)
        ax = fig.add_subplot(1, 1, 1, projection=self.proj)
        ax.set_extent(self.extent, crs=self.proj)
        if not self.args.test:  # pragma: no cover
            ax.coastlines(resolution='50m')
            ax.add_feature(cartopy.feature.LAND)
            ax.add_feature(cartopy.feature.OCEAN)
            ax.add_feature(cartopy.feature.COASTLINE)
            ax.add_feature(cartopy.feature.BORDERS, linestyle=':')
            ax.add_feature(cartopy.feature.LAKES, alpha=0.5)
            ax.add_feature(cartopy.feature.RIVERS)
        self.ax = ax
        # Figure out the scaling factor between degrees and pixels. Many config values are given
        # in pixels but need to be converted to degrees for plotting.
        m = self.proj._as_mpl_transform(axes=self.ax)
        y0 = m.transform_point((self.extent[0], self.extent[2])).tolist()[1]
        y1 = m.transform_point((self.extent[0], self.extent[2] + 1)).tolist()[1]
        self.scaling_factor = 1.0 / (y1 - y0)
        # So 1 px = self.scaling_factor * 1°
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self.args.title:
            plt.title(self.args.title)
        format = self.args.output.suffix.replace('.', '').lower()
        if format == 'jpg':
            mplfname = self.args.output.parent / '{}.png'.format(self.args.output.stem)
            plt.savefig(str(mplfname), bbox_inches="tight")
            img = Image.open(str(mplfname)).convert('RGB')
            img.save(str(self.args.output), optimize=True, quality=95)
            mplfname.unlink()
        else:
            plt.savefig(str(self.args.output), bbox_inches="tight")

        plt.close()

    @staticmethod
    def add_options(parser, help_suffix):
        for direction in ['left', 'right', 'top', 'bottom']:
            parser.add_argument(
                '--padding-{}'.format(direction),
                help="{} padding of the map in degrees. {}".format(
                    direction.capitalize(), help_suffix),
                type=int,
                default=1,
            )
        parser.add_argument(
            '--width',
            help="Width of the figure in inches. {}".format(help_suffix),
            type=float,
            default=6.4,
        )
        parser.add_argument(
            '--height',
            help="Height of the figure in inches. {}".format(help_suffix),
            type=float,
            default=4.8,
        )
        parser.add_argument(
            '--dpi',
            help="Pixel density of the figure. {}".format(help_suffix),
            type=float,
            default=100.0,
        )

    def _lonlat(self, language):
        lat, lon = language.lat, language.lon
        if self.central_longitude:
            lon = lon - self.central_longitude
            if lon < -180:
                lon += 360
        return lon, lat

    def plot_language_marker(self,
                             language,
                             text=None,
                             marker_kw=None,
                             text_kw=None,
                             text_offset=(-0.075, -0.075)):  # pragma: no cover
        lon, lat = self._lonlat(language)
        kw = dict(color='red', markersize=self.args.markersize, zorder=4, marker='o')
        if marker_kw:
            kw.update(marker_kw)
        self.ax.plot(lon, lat, **kw)

        if text:
            kw = dict(fontsize=10, zorder=5)
            if text_kw:
                kw.update(text_kw)
            self.ax.text(lon + text_offset[0], lat + text_offset[1], text, **kw)

    def add_language(self, language, values, colormaps):
        s, angle = 0, 360.0 / len(values)
        lon, lat = self._lonlat(language)
        for pid, vals in values.items():
            self.ax.add_patch(Wedge(
                [lon, lat],
                self.args.markersize * self.scaling_factor / 2.0,
                s,
                s + angle,
                facecolor=colormaps[pid](vals[0].v),
                edgecolor="black",
                label=language.name,
            ))
            s += angle
        if self.args.language_labels:
            self.ax.text(
                lon + self.args.markersize * self.scaling_factor + 3 * self.scaling_factor,
                lat, language.name, fontsize='small')

    def add_legend(self, parameters, colormaps):
        def wrapped_label(s):
            return '\n'.join(textwrap.wrap(s, width=20))

        handles = []
        s, angle = 0, 360.0 / len(parameters)
        for pid, parameter in parameters.items():
            handles.append(Wedge(
                [-100, -100],
                self.args.markersize,
                s,
                s + angle,
                facecolor="None",
                edgecolor="black",
                label=wrapped_label(parameter.name),
            ))
            if isinstance(parameter.domain, tuple):
                cbar = plt.colorbar(
                    colormaps[pid].scalar_mappable(),
                    ax=self.ax,
                    aspect=80,
                    label=parameter.name,
                    shrink=0.5,
                    pad=0.05,
                    location="bottom",
                )
                cbar.set_ticks([0.0, 1.0])
                cbar.set_ticklabels([
                    str(round(parameter.domain[0], 2)),
                    str(round(parameter.domain[1], 2))])
            else:
                for v, label in parameter.domain.items():
                    handles.append(Wedge(
                        [-100, -100],
                        self.args.markersize,
                        s,
                        s + angle,
                        facecolor=colormaps[pid](v),
                        edgecolor="black",
                        label=wrapped_label(label),
                    ))
            s += angle
        self.ax.legend(
            bbox_to_anchor=(1, 1),
            handles=handles,
            loc='upper left',
            handler_map={Wedge: HandleWedge()})
